/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
#include "caosdb/message_code.h"   // for MessageCode
#include "caosdb/status_code.h"    // for StatusCode
#include "ccaosdb.h"               // for caosdb_info_version_info, caosdb_...
#include "gtest/gtest-message.h"   // for Message
#include "gtest/gtest-test-part.h" // for TestPartResult
#include "gtest/gtest_pred_impl.h" // for Test, EXPECT_EQ, SuiteApiResolver
#include <memory>                  // for allocator, unique_ptr

TEST(test_ccaosdb, connection_ssl_authentication_success) {

  caosdb_connection_connection connection;
  caosdb_connection_connection_manager_get_default_connection(&connection);

  caosdb_info_version_info version_info;
  caosdb_connection_get_version_info(&version_info, &connection);

  auto major = caosdb_constants_COMPATIBLE_SERVER_VERSION_MAJOR();
  auto minor = caosdb_constants_COMPATIBLE_SERVER_VERSION_MINOR();
  const auto *const pre_release =
      caosdb_constants_COMPATIBLE_SERVER_VERSION_PRE_RELEASE();

  EXPECT_EQ(major, version_info.major);
  EXPECT_EQ(minor, version_info.minor);
  EXPECT_STREQ(pre_release, version_info.pre_release);
}

TEST(test_ccaosdb, test_count_query) {

  caosdb_connection_connection connection;
  int return_code(
    caosdb_connection_connection_manager_get_default_connection(&connection));
  EXPECT_EQ(return_code, 0);

  caosdb_transaction_transaction transaction;
  return_code =
    caosdb_connection_connection_create_transaction(&connection, &transaction);
  EXPECT_EQ(return_code, 0);

  // COUNT query with an empty result
  return_code = caosdb_transaction_transaction_query(&transaction,
                                                     "COUNT ENTITY WITH id=-1");
  EXPECT_EQ(return_code, caosdb::StatusCode::GO_ON);

  return_code = caosdb_transaction_transaction_execute(&transaction);
  EXPECT_EQ(return_code, 0);

  long dummy(-1);       // NOLINT
  long *count = &dummy; // NOLINT
  return_code =
    caosdb_transaction_transaction_get_count_result(&transaction, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 0);

  return_code = caosdb_transaction_delete_transaction(&transaction);
  EXPECT_EQ(return_code, 0);
}

TEST(test_ccaosdb, test_query) {
  caosdb_connection_connection connection;
  int return_code(
    caosdb_connection_connection_manager_get_default_connection(&connection));
  EXPECT_EQ(return_code, 0);

  caosdb_transaction_transaction transaction;
  return_code =
    caosdb_connection_connection_create_transaction(&connection, &transaction);
  EXPECT_EQ(return_code, 0);

  return_code = caosdb_transaction_transaction_query(&transaction,
                                                     "FIND ENTITY WITH id=-1");
  EXPECT_EQ(return_code, caosdb::StatusCode::GO_ON);

  return_code = caosdb_transaction_transaction_execute(&transaction);
  EXPECT_EQ(return_code, 0);

  caosdb_transaction_result_set result_set;
  return_code =
    caosdb_transaction_transaction_get_result_set(&transaction, &result_set);
  EXPECT_EQ(return_code, 0);

  int dummy(-1);
  int *count = &dummy; // NOLINT
  return_code = caosdb_transaction_result_set_size(&result_set, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 0);

  return_code = caosdb_transaction_delete_transaction(&transaction);
  EXPECT_EQ(return_code, 0);
}

TEST(test_ccaosdb, test_single_id_retrieve) {
  caosdb_connection_connection connection;
  int return_code(
    caosdb_connection_connection_manager_get_default_connection(&connection));
  EXPECT_EQ(return_code, 0);

  caosdb_transaction_transaction transaction;
  return_code =
    caosdb_connection_connection_create_transaction(&connection, &transaction);
  EXPECT_EQ(return_code, 0);

  return_code =
    caosdb_transaction_transaction_retrieve_by_id(&transaction, "21");
  EXPECT_EQ(return_code, caosdb::StatusCode::GO_ON);

  return_code = caosdb_transaction_transaction_execute(&transaction);
  EXPECT_EQ(return_code, 0);

  caosdb_transaction_result_set result_set;
  return_code =
      caosdb_transaction_transaction_get_result_set(&transaction, &result_set);
  EXPECT_EQ(return_code, 0);

  int dummy(-1);
  int *count = &dummy; // NOLINT
  return_code = caosdb_transaction_result_set_size(&result_set, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 1);

  caosdb_entity_entity entity;
  return_code = caosdb_transaction_result_set_at(&result_set, &entity, 0);
  EXPECT_EQ(return_code, 0);

  return_code = caosdb_entity_entity_get_errors_size(&entity, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 0);

  char out[255] = {"255"}; // NOLINT
  return_code = caosdb_entity_entity_get_name(&entity, out);
  EXPECT_EQ(return_code, 0);
  EXPECT_STREQ(out, "unit");

  return_code = caosdb_transaction_delete_transaction(&transaction);
  EXPECT_EQ(return_code, 0);
}

TEST(test_ccaosdb, test_multi_id_retrieve) {
  caosdb_connection_connection connection;
  int return_code(
    caosdb_connection_connection_manager_get_default_connection(&connection));
  EXPECT_EQ(return_code, 0);

  caosdb_transaction_transaction transaction;
  return_code =
    caosdb_connection_connection_create_transaction(&connection, &transaction);
  EXPECT_EQ(return_code, 0);

  const char *ids[] = {"20", "21", "22"}; // NOLINT
  return_code =
    caosdb_transaction_transaction_retrieve_by_ids(&transaction, ids, 3);
  EXPECT_EQ(return_code, caosdb::StatusCode::GO_ON);

  return_code = caosdb_transaction_transaction_execute(&transaction);
  // Should have an error since entity 22 doesn't exist
  EXPECT_TRUE((return_code > 0));

  caosdb_transaction_result_set result_set;
  return_code =
    caosdb_transaction_transaction_get_result_set(&transaction, &result_set);
  EXPECT_EQ(return_code, 0);

  int dummy(-1);
  int *count = &dummy; // NOLINT
  return_code = caosdb_transaction_result_set_size(&result_set, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 3);

  caosdb_entity_entity entity;
  return_code = caosdb_transaction_result_set_at(&result_set, &entity, 1);
  EXPECT_EQ(return_code, 0);

  return_code = caosdb_entity_entity_get_errors_size(&entity, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 0);

  char out[255] = {"255"}; // NOLINT
  return_code = caosdb_entity_entity_get_name(&entity, out);
  EXPECT_EQ(return_code, 0);
  EXPECT_STREQ(out, "unit");

  caosdb_entity_entity other_entity;
  return_code = caosdb_transaction_result_set_at(&result_set, &other_entity, 2);
  EXPECT_EQ(return_code, 0);
  return_code = caosdb_entity_entity_get_errors_size(&other_entity, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, 1);

  caosdb_entity_message err_msg;
  return_code = caosdb_entity_entity_get_error(&other_entity, &err_msg, 0);
  EXPECT_EQ(return_code, 0);

  return_code = caosdb_entity_message_get_code(&err_msg, count);
  EXPECT_EQ(return_code, 0);
  EXPECT_EQ(*count, caosdb::entity::MessageCode::ENTITY_DOES_NOT_EXIST);

  return_code = caosdb_transaction_delete_transaction(&transaction);
  EXPECT_EQ(return_code, 0);
}
